(* konvertieren von Pixelbildern nach Encapsulted Postscript

    J. Rathlev, D-24222 Schwentinental, (kontakt(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1   : Jan. 2006
   Vers. 1.1 : Mai 2007, auch Png-Format, Strings als Ressourcen
         1.2 : Jan. 2008, Mehrsprachigkeit ber GnuGetText
         2.0 : Mrz 2018, Delphi 10
   *)

unit Img2EpsMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.Buttons,
  Vcl.ComCtrls, Vcl.ExtCtrls, WinApiUtils;

const
  ProgName = 'Image to EPS';
  Vers = ' Vers. 2.1';
  CopRgt = ' 2012-2020 Dr. J. Rathlev, D-24222 Schwentinental';
  EMailAdr = 'kontakt(a)rathlev-home.de';

  ImageFormats = '0.bmp;1.jpg;2.png;3.pcx;4.tga;5.tif;6.gif';

type
  TfrmPic2Eps = class(TForm)
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    btnImgDir: TSpeedButton;
    btnImgFiles: TSpeedButton;
    OpenDialog: TOpenDialog;
    btnEpsDir: TSpeedButton;
    btnQuit: TBitBtn;
    btnInfo: TSpeedButton;
    gpbOptions: TGroupBox;
    cbCompress: TCheckBox;
    cbInvert: TCheckBox;
    cbColor: TCheckBox;
    lblStatus: TLabel;
    Panel1: TPanel;
    rgRotate: TRadioGroup;
    btnStart: TBitBtn;
    edtImgDir: TComboBox;
    edtImgFiles: TComboBox;
    edtEpsDir: TComboBox;
    procedure btnImgDirClick(Sender: TObject);
    procedure btnEpsDirClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnQuitClick(Sender: TObject);
    procedure btnImgFilesClick(Sender: TObject);
    procedure btnConvertClick(Sender: TObject);
    procedure btnInfoClick(Sender: TObject);
    procedure cbColorClick(Sender: TObject);
  private
    { Private-Deklarationen }
    ProgVersName,
    ProgVersDate,
    AppPath,
    DefPath,TmpPath,
    Ininame         : string;
    function PicToEps (ImgName,EpsPath : string) : boolean;
  public
    { Public-Deklarationen }
  end;

var
  frmPic2Eps: TfrmPic2Eps;

implementation

{$R *.dfm}

uses System.IniFiles, Vcl.Imaging.jpeg, Vcl.Imaging.PngImage, Vcl.Imaging.GifImg,
  Winapi.ShlObj, System.StrUtils, WinUtils, WinShell, GnuGettext, LangUtils, InitProg,
  StringUtils, FileUtils, ShellDirDlg, Eps, TiffImage, TargaImage, PcxImage;

const
  IniExt = 'ini';

  (* INI-Sektionen *)
  CfGSekt = 'Config';
  SourceSekt = 'Source';
  DestSekt = 'Dest';
  FileSekt = 'Files';

  (* INI-Variablen *)
  iniComp  = 'Compress';
  iniColor = 'Color';
  iniInv   = 'Invert';
  iniRot   = 'Rotate';

procedure TfrmPic2Eps.FormCreate(Sender: TObject);
var
  IniFile  : TIniFile;
begin
  TranslateComponent(self);
  InitPaths(AppPath,DefPath);
  TmpPath:=IncludeTrailingPathDelimiter(GetEnvironmentVariable('TEMP'));
// Versions-Info
  InitVersion(ProgName,Vers,CopRgt,3,3,ProgVersName,ProgVersDate);
  IniName:=Erweiter(AppPath,PrgName,IniExt);
  IniFile:=TIniFile.Create(IniName);
  with IniFile do begin
    cbCompress.Checked:=ReadBool(CfgSekt,iniComp,true);
    cbColor.Checked:=ReadBool(CfgSekt,iniColor,false);
    with cbInvert do begin
      Checked:=ReadBool(CfgSekt,iniInv,false);
      Enabled:=not cbColor.Checked;
      end;
    rgRotate.Itemindex:=Readinteger(CfgSekt,iniRot,0);
    Free;
    end;
  LoadHistory(IniName,SourceSekt,edtImgDir);
  edtImgDir.ItemIndex:=0;
  LoadHistory(IniName,FileSekt,edtImgFiles);
  edtImgFiles.ItemIndex:=0;
  LoadHistory(IniName,DestSekt,edtEpsDir);
  edtEpsDir.ItemIndex:=0;
  lblStatus.Caption:='';
  end;

procedure TfrmPic2Eps.FormDestroy(Sender: TObject);
var
  IniFile  : TIniFile;
begin
  IniFile:=TIniFile.Create(IniName);
  with IniFile do begin
    WriteBool(CfgSekt,iniComp,cbCompress.Checked);
    WriteBool(CfgSekt,iniColor,cbColor.Checked);
    WriteBool(CfgSekt,iniInv,cbInvert.Checked);
    WriteInteger(CfgSekt,iniRot,rgRotate.Itemindex);
    Free;
    end;
  SaveHistory(IniName,SourceSekt,true,edtImgDir);
  SaveHistory(IniName,FileSekt,true,edtImgFiles);
  SaveHistory(IniName,DestSekt,true,edtEpsDir);
  end;

procedure TfrmPic2Eps.btnImgDirClick(Sender: TObject);
var
  s : string;
begin
  s:=edtImgDir.Text;
  if length(s)=0 then s:=DefPath;
  if ShellDirDialog.Execute (_('Folder to image files'),true,true,false,'',s) then begin
    with edtImgDir do begin
      Text:=s; AddToHistory(edtImgDir,s);
      end;
    with edtEpsDir do if length(Text)=0 then begin
      Text:=s; AddToHistory(edtEpsDir,s);
      end;
    with edtImgFiles do if length(Text)=0 then begin
      Text:='*.jpg'; AddToHistory(edtImgFiles,Text);
      end;
    end;
  end;

procedure TfrmPic2Eps.btnEpsDirClick(Sender: TObject);
var
  s : string;
begin
  s:=edtEpsDir.Text;
  if length(s)=0 then s:=edtImgDir.Text;
  if length(s)=0 then s:=DefPath;
  if ShellDirDialog.Execute (_('Folder to EPS output'),true,true,false,edtImgDir.Text,s) then begin
    edtEpsDir.Text:=s; AddToHistory(edtEpsDir,s);
    end;
  end;

procedure TfrmPic2Eps.btnQuitClick(Sender: TObject);
begin
  Close;
  end;

procedure TfrmPic2Eps.btnImgFilesClick(Sender: TObject);
var
  i : integer;
  s : string;
begin
  with OpenDialog do begin
    InitialDir:=edtImgDir.Text;
    Options:=[ofHideReadOnly,ofAllowMultiSelect,ofPathMustExist,ofFileMustExist,ofEnableSizing];
    Filename:='';
    Title:=_('Select image files');
    if Execute then begin
      s:=ExtractFilePath(Files.Strings[0]);
      with edtImgDir do if s<>Text then begin
        Text:=s; AddToHistory(edtImgDir,s);
        end;
      with Files do begin
        s:='';
        for i:=0 to Count-1 do s:=s+ExtractFilename(Strings[i])+';';
        end;
      delete (s,length(s),1);
      edtImgFiles.Text:=s; AddToHistory(edtImgFiles,s);
      end;
    end;
  end;

function TfrmPic2Eps.PicToEps (ImgName,EpsPath : string) : boolean;
var
  bp    : TPicture;
  bTif  : TTiffImage;
  bTga  : TTarga;
  bPcx  : TPcxImage;
  bPng  : TPngImage;
  bGif  : TGifImage;
  bm    : TBitMap;
  eg    : TEpsGraphic;
  ext   : string;
  i,j   : integer;
  en    : string;
  ok    : boolean;
begin
  Result:=false;
  ext:=AnsiLowerCase(ExtractFileExt(ImgName));
  i:=AnsiPos(ext,ImageFormats);
  if i=0 then exit;   // unknown format
  en:=IncludeTrailingPathDelimiter(EpsPath)+ExtractFileName(ChangeFileExt(ImgName,EpsExt));
  j:=emNone;
  if cbCompress.Checked then j:=j or emCompress;
  if cbColor.Checked then j:=j or emColor
  else if cbInvert.Checked then j:=j or emInvert;
  eg:=TEpsGraphic.Create(j,TRotate(rgRotate.ItemIndex));
  eg.Creator:=ProgVersName;
  eg.ImageName:=ImgName;
  bm:=TBitMap.Create;
  i:=ord(ImageFormats[i-1])-48;
  ok:=false;
  case i of
  0,1 : begin  // bmp and jpg
        bp:=TPicture.Create;
        with bp do begin
          LoadFromFile(ImgName);
          ok:=not Graphic.Empty;
          end;
        if ok then with bm do begin
          Assign(bp.Graphic);
          if PixelFormat=pf16bit then PixelFormat:=pf24bit;
          end;
        bp.Free;
        end;
  2 :   begin // png
        bPng:=TPngImage.Create;
        with bPng do begin
          LoadFromFile(ImgName);
          ok:=not Empty;
          end;
        if ok then with bm do begin
          bPng.RemoveTransparency;
          Assign(bPng);
          end;
        bPng.Free;
        end;
  3 :   begin // Pcx
//        bPcx:=TPcxBitmap.Create;
        bPcx:=TPcxImage.Create;
        with bPcx do begin
          LoadFromFile(ImgName);
          ok:=not Empty;
          end;
        if ok then with bm do begin
          Assign(bPcx);
          end;
        bPcx.Free;
        end;
  4 :   begin // Tga
        bTga:=TTarga.Create;
        with bTga do begin
          LoadFromFile(ImgName);
          ok:=not Empty;
          end;
        if ok then with bm do begin
          Assign(bTga.Bitmap);
          if PixelFormat=pf32bit then PixelFormat:=pf24bit;
          end;
        bTga.Free;
        end;
  5 :   begin // Tiff
        bTif:=TTiffImage.Create;
        with bTif do begin
          LoadFromTifFileEx(ImgName);
          ok:=not Empty;
          SaveToTifFile(ExtractFilePath(ImgName)+'test.tif',true);
          end;
        if ok then with bm do begin
          Assign(bTif);
          end;
        bTif.Free;
        end;
  6 : begin //Gif
      bGif:=TGifImage.Create;
      with bGif do begin
        LoadFromFile(ImgName);
        ok:=not Empty;
        end;
      if ok then with bm do begin
        Assign(bGif.Bitmap);
        PixelFormat:=pf8bit;
//        Canvas.StretchDraw (Rect(0,Height-1,Width-1,0),bGif);  // kippen
        end;
      bGif.Free;
      end;
    end;
  if ok then Result:=eg.WriteToFile(en,bm);
  bm.Free; eg.Free;
  end;

procedure TfrmPic2Eps.btnConvertClick(Sender: TObject);
var
  FListe     : TStringList;
  n,i        : integer;
  fn,s,t     : string;
  ok         : boolean;

  procedure ScanDir (Dir,Mask : string);
  var
    DirInfo    : TSearchRec;
    Findresult : integer;
  begin
    FindResult:=FindFirst (Erweiter(Dir,Mask,''),faArchive+faReadOnly+faHidden+faSysfile,DirInfo);
    while FindResult=0 do with DirInfo do begin
      if NotSpecialDir(Name) then FListe.Add(Erweiter(Dir,Name,''));
      FindResult:=FindNext (DirInfo);
      end;
    FindClose(DirInfo);
    end;

begin
  t:=edtImgFiles.Text;
  ok:=length(t)>0;
  if ok and not DirectoryExists(edtEpsDir.Text) then begin
    ok:=ConfirmDialog(Caption,Format(_('EPS-Directory: '+sLineBreak+' %s '+sLineBreak+
                    'does not exist - Create?'),[edtEpsDir.Text]));
    if ok then ok:=ForceDirectories(edtEpsDir.Text);
    end;
  if ok then begin
    FListe:=TStringList.Create;
    n:=0;
    repeat
      s:=ReadNxtStr (t,';');
      FListe.Clear;
      ScanDir (edtImgDir.Text,s);
      with FListe do if Count>0 then for i:=0 to Count-1 do begin
        fn:=StripPath(Strings[i],44);
        lblStatus.Caption:=Format(_('File: %s'),[StripPath(fn,50)]);
        Application.ProcessMessages;
        if PicToEPS (Strings[i],edtEpsDir.Text) then inc(n)
        else ErrorDialog(Caption,Format(_('Error converting:'+sLineBreak+'%s'),[fn]));
        end;
      until length(t)=0;
    FListe.Free;
    if n>0 then begin
      if n>1 then lblStatus.Caption:=Format(_('%u files converted!'),[n])
      else lblStatus.Caption:=Format(_('%s converted!'),[StripPath(fn,40)]);
      end
    else lblStatus.Caption:=_('No matching files found!');
    end
  end;

procedure TfrmPic2Eps.btnInfoClick(Sender: TObject);
begin
  InfoDialog(ProgName,Format(_('%s - %s'+sLineBreak+'%s'+sLineBreak+'E-Mail: %s'),
     [ProgVersName,ProgVersDate,CopRgt,EmailAdr]));
  end;

procedure TfrmPic2Eps.cbColorClick(Sender: TObject);
begin
  cbInvert.Enabled:=not cbColor.Checked;
  end;

end.
