(* Convert png images to bmp images

    Dr. J. Rathlev, D-24222 Schwentinental

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   J. Rathlev, July 2015
   last modified: March 2018
   *)

unit PngToBmpMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.Buttons,
  Vcl.ExtCtrls, Vcl.ExtDlgs;

const
  ProgName = 'Convert PNG to BMP';
  Vers = ' Vers. 1.2';
  CopRgt = ' 2015-2020 J. Rathlev, D-24222 Schwentinental';
  EMailAdr = 'kontakt(a)rathlev-home.de';

  phDir = '%SubDir%';

type
  TMainForm = class(TForm)
    paTop: TPanel;
    Label1: TLabel;
    btnImgDir: TSpeedButton;
    Label2: TLabel;
    btnImgFiles: TSpeedButton;
    Label3: TLabel;
    btnBmpDir: TSpeedButton;
    paMain: TPanel;
    paBottom: TPanel;
    btnInfo: TSpeedButton;
    btnQuit: TBitBtn;
    OpenPictureDialog: TOpenPictureDialog;
    paImage: TPanel;
    PaintBox: TPaintBox;
    bbBgColor: TBitBtn;
    ColorDialog: TColorDialog;
    bbConvert: TBitBtn;
    lblStatus: TLabel;
    laSample: TLabel;
    gbSettings: TGroupBox;
    Label4: TLabel;
    rb8Bit: TRadioButton;
    rb24Bit: TRadioButton;
    rb32Bit: TRadioButton;
    lblStatusHeader: TLabel;
    btnSubdir: TSpeedButton;
    edtImgDir: TComboBox;
    edtImgFiles: TComboBox;
    edtBmpDir: TComboBox;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnInfoClick(Sender: TObject);
    procedure btnQuitClick(Sender: TObject);
    procedure btnImgDirClick(Sender: TObject);
    procedure btnImgFilesClick(Sender: TObject);
    procedure btnBmpDirClick(Sender: TObject);
    procedure PaintBoxPaint(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure bbBgColorClick(Sender: TObject);
    procedure bbConvertClick(Sender: TObject);
    procedure edtImgDirCloseUp(Sender: TObject);
    procedure edtImgFilesCloseUp(Sender: TObject);
    procedure rb8BitClick(Sender: TObject);
    procedure rb24BitClick(Sender: TObject);
    procedure rb32BitClick(Sender: TObject);
    procedure btnSubdirClick(Sender: TObject);
  private
    { Private-Deklarationen }
    ProgVersName,
    ProgVersDate,
    AppPath,
    DefPath,TmpPath,
    Ininame         : string;
    FImage          : TBitMap;
    PixelFormat     : TPixelFormat;
    function ReplDirPh (const APath : string) : string;
    function LoadPngImage (const FName : string; Bitmap :  TBitmap) : boolean;
    procedure LoadImage;
  public
    { Public-Deklarationen }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

uses System.IniFiles, Vcl.Imaging.PngImage, Vcl.FileCtrl, System.StrUtils,
  StringUtils, WinUtils, FileUtils, LangUtils, InitProg, GnuGetText, ShellDirDlg;

const
  IniExt = 'ini';

  (* INI-Sektionen *)
  CfGSekt = 'Config';
  SourceSekt = 'Source';
  DestSekt = 'Dest';
  FileSekt = 'Files';

  iniTop      = 'Top';
  iniLeft     = 'Left';
  iniWdt      = 'Width';
  iniHgt      = 'Height';
  iniBgColor  = 'BackgroundColor';
  iniFormat   = 'PixelFormat';

procedure TMainForm.FormCreate(Sender: TObject);
var
  IniFile  : TIniFile;
begin
  TranslateComponent (self);
  InitPaths(AppPath,DefPath);
  TmpPath:=IncludeTrailingPathDelimiter(GetEnvironmentVariable('TEMP'));
// Versions-Info
  InitVersion(ProgName,Vers,CopRgt,3,3,ProgVersName,ProgVersDate);
  IniName:=Erweiter(AppPath,PrgName,IniExt);
  IniFile:=TIniFile.Create(IniName);
  with IniFile do begin
    Top:=ReadInteger(CfgSekt,iniTop,Top);
    Left:=ReadInteger(CfgSekt,iniLeft,Left);
    ClientWidth:=ReadInteger(CfgSekt,iniWdt,ClientWidth);
    ClientHeight:=ReadInteger(CfgSekt,iniHgt,ClientHeight);
    paImage.Color:=ReadInteger(CfgSekt,iniBgColor,clGray);
    PixelFormat:=TPixelFormat(ReadInteger(CfgSekt,iniFormat,integer(pf24bit)));
    Free;
    end;
  LoadHistory(IniName,SourceSekt,edtImgDir);
  edtImgDir.ItemIndex:=0;
  LoadHistory(IniName,FileSekt,edtImgFiles);
  edtImgFiles.ItemIndex:=0;
  LoadHistory(IniName,DestSekt,edtBmpDir);
  edtBmpDir.ItemIndex:=0;
  FImage:=TBitMap.Create;
  end;

procedure TMainForm.FormDestroy(Sender: TObject);
var
  IniFile  : TIniFile;
begin
  IniFile:=TIniFile.Create(IniName);
  with IniFile do begin
    WriteInteger(CfgSekt,iniTop,Top);
    WriteInteger(CfgSekt,iniLeft,Left);
    WriteInteger(CfgSekt,iniWdt,ClientWidth);
    WriteInteger(CfgSekt,iniHgt,ClientHeight);
    WriteInteger(CfgSekt,iniBgColor,paImage.Color);
    WriteInteger(CfgSekt,iniFormat,integer(PixelFormat));
    Free;
    end;
  SaveHistory(IniName,SourceSekt,true,edtImgDir);
  SaveHistory(IniName,FileSekt,true,edtImgFiles);
  SaveHistory(IniName,DestSekt,true,edtBmpDir);
  FImage.Free;
  end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  LoadImage;
  end;

procedure TMainForm.PaintBoxPaint(Sender: TObject);
var
  n : integer;
begin
  if assigned(FImage) then with PaintBox,Canvas do begin
    if Width>Height then n:=Height else n:=Width;
    StretchDraw(Rect(0,0,n,n),FImage);
    end;
  end;

procedure TMainForm.rb24BitClick(Sender: TObject);
begin
  PixelFormat:=pf24bit;
  end;

procedure TMainForm.rb32BitClick(Sender: TObject);
begin
  PixelFormat:=pf32bit;
  end;

procedure TMainForm.rb8BitClick(Sender: TObject);
begin
  PixelFormat:=pf8bit;
  end;

procedure TMainForm.btnQuitClick(Sender: TObject);
begin
  Close;
  end;

procedure TMainForm.btnSubdirClick(Sender: TObject);
begin
  with edtBmpDir do if Pos(phDir,Text)=0 then Text:=SetDirName(Text)+phDir;
  end;

function TMainForm.ReplDirPh (const APath : string) : string;
begin
  if AnsiContainsText(APath,phDir) then
    Result:=AnsiReplaceText(APath,phDir,ExtractFilename(ExcludeTrailingPathDelimiter(edtImgDir.Text)))
  else Result:=APath;
  end;

procedure TMainForm.edtImgDirCloseUp(Sender: TObject);
var
  n : integer;
  s : string;
begin
  with edtImgDir do Text:=Items[ItemIndex];
  with edtBmpDir do if not AnsiContainsStr(Text,phDir) then begin
    n:=Items.IndexOf(edtImgDir.Text);
    if n>=0 then Text:=Items[n];
    end;
  LoadImage;
  end;

procedure TMainForm.edtImgFilesCloseUp(Sender: TObject);
begin
  LoadImage;
  end;

procedure TMainForm.btnInfoClick(Sender: TObject);
begin
  InfoDialog(ProgName,Format(_('%s - %s'+sLineBreak+'%s'+sLineBreak+'E-Mail: %s'),
             [ProgVersName,ProgVersDate,CopRgt,EmailAdr]));
  end;

procedure TMainForm.bbBgColorClick(Sender: TObject);
begin
  with ColorDialog do begin
    Color:=paImage.Color;
    if Execute then begin
      paImage.Color:=Color;
      LoadImage;
      end;
    end;
  end;

procedure TMainForm.btnBmpDirClick(Sender: TObject);
var
  s : string;
begin
  s:=edtBmpDir.Text;
  if length(s)=0 then s:=edtImgDir.Text;
  if length(s)=0 then s:=DefPath;
  if ShellDirDialog.Execute (_('Folder for BMP images'),true,true,false,edtImgDir.Text,s) then begin
    edtBmpDir.Text:=s;
    AddToHistory(edtBmpDir,s);
    end;
  end;

procedure TMainForm.btnImgDirClick(Sender: TObject);
var
  s : string;
  n : integer;
begin
  s:=edtImgDir.Text;
  if length(s)=0 then s:=DefPath;
  if ShellDirDialog.Execute (_('Folder with PNG images'),true,true,false,'',s) then begin
    edtImgDir.Text:=s; AddToHistory(edtImgDir,s);
    with edtBmpDir do if not AnsiContainsStr(Text,phDir) then begin
      Text:=s; AddToHistory(edtBmpDir,s);
      end;
    with edtImgFiles do if length(Text)=0 then begin
      Text:='*.png'; AddToHistory(edtImgFiles,Text);
      end;
    LoadImage;
    end;
  end;

procedure TMainForm.btnImgFilesClick(Sender: TObject);
var
  s : string;
  i : integer;
begin
  with OpenPictureDialog do begin
    InitialDir:=edtImgDir.Text;
    Options:=[ofHideReadOnly,ofAllowMultiSelect,ofPathMustExist,ofFileMustExist,ofEnableSizing];
    Filename:='';
    Title:=_('Select PNG images');
    if Execute then begin
      s:=ExtractFilePath(Files.Strings[0]);
      with edtImgDir do if s<>Text then begin
        Text:=s; AddToHistory(edtImgDir,s);
        end;
      with Files do begin
        s:='';
        for i:=0 to Count-1 do s:=s+ExtractFilename(Strings[i])+';';
        end;
      Delete (s,length(s),1);
      edtImgFiles.Text:=s; AddToHistory(edtImgFiles,s);
      LoadImage;
      end;
    end;
  end;

function TMainForm.LoadPngImage (const FName : string; Bitmap : TBitmap) : boolean;
var
  bp   : TPngImage;
begin
  Result:=false;
  if FileExists(FName) then begin
    bp:=TPngObject.Create;
    bp.LoadFromFile(FName);
    if not bp.Empty then with Bitmap do begin
      Width:=bp.Width; Height:=bp.Height;
      with Canvas do begin
        Brush.Color:=paImage.Color;
        FillRect(Rect(0,0,Width,Height));
        Draw(0,0,bp);
        end;
//      Assign(bp);
      Transparent:=true;
      TransParentColor:=paImage.Color;
      Result:=true;
      end;
    bp.Free;
    end;
  end;

procedure TMainForm.LoadImage;
var
  DirInfo    : TSearchRec;
  FindResult : integer;
  s,t  : string;
begin
  t:=edtImgFiles.Text;
  s:=ReadNxtStr (t,';');
  FindResult:=FindFirst (Erweiter(edtImgDir.Text,s,''),faArchive+faReadOnly+faHidden+faSysfile,DirInfo);
  if FindResult=0 then with DirInfo do begin
    LoadPngImage(Erweiter(edtImgDir.Text,Name,''),FImage);
    laSample.Caption:=_('Sample: ')+DirInfo.Name;
    end
  else with FImage do begin
    Width:=0; Height:=0;
    laSample.Caption:=_('No matching image found!');
    end;
  FindClose(DirInfo);
  lblStatusHeader.Caption:=''; lblStatus.Caption:='';
  PaintBox.Invalidate;
  end;

procedure TMainForm.bbConvertClick(Sender: TObject);
var
  FListe     : TStringList;
  i          : integer;
  s,t,sd     : string;
  ok         : boolean;
  bm         : TBitmap;

  procedure ScanDir (Dir,Mask : string);
  var
    DirInfo    : TSearchRec;
    Findresult : integer;
  begin
    FindResult:=FindFirst (Erweiter(Dir,Mask,''),faArchive+faReadOnly+faHidden+faSysfile,DirInfo);
    while FindResult=0 do with DirInfo do begin
      if NotSpecialDir(Name) then FListe.Add(Erweiter(Dir,Name,''));
      FindResult:=FindNext (DirInfo);
      end;
    FindClose(DirInfo);
    end;

begin
  t:=edtImgFiles.Text;
  ok:=length(t)>0;
  sd:=ReplDirPh(edtBmpDir.Text);
  if ok and not DirectoryExists(sd) then begin
    ok:=ConfirmDialog(Caption,Format(_('BMP directory: '+sLineBreak+' %s '+sLineBreak+
                    'does not exist - Create?'),[sd]));
    if ok then ok:=ForceDirectories(sd);
    end;
  if ok then begin
    FListe:=TStringList.Create;
    bm:=TBitmap.Create;
    repeat
      s:=ReadNxtStr (t,';');
      ScanDir (edtImgDir.Text,s);
      until length(t)=0;
    lblStatus.Caption:='';
    with FListe do if Count>0 then begin
      lblStatusHeader.Caption:=_('Converting:');
      for i:=0 to Count-1 do begin
        with lblStatus do Caption:=MinimizeName(Strings[i],Canvas,Width);
        LoadPngImage(Strings[i],bm);
        bm.PixelFormat:=PixelFormat;
        bm.SaveToFile(SetDirName(sd)+NewExt(ExtractFilename(Strings[i]),'.bmp'));
        Application.ProcessMessages;
        end;
      if Count>1 then begin
        lblStatusHeader.Caption:=Format(_('%u images converted!'),[Count]);
        lblStatus.Caption:='';
        end
      else lblStatusHeader.Caption:=_('Converted:');
      end
    else lblStatusHeader.Caption:=_('No matching images found!');
    FListe.Free; bm.Free;
    end;
  end;

end.
